
### ###############################################################################
### # Filename: mc.magic.tcl                                                      #
### #                                                                             #
### # Description:                                                                #
### #      /                                      #
### ###############################################################################

###  
proc mc:magic:dmgmod {dmg id} {
  set mag [expr 0.5 * [mc:round:var mbonus $id] * ([mc:round:var magatt $id] + [mc:round:var magprt $id]) - 20]
  if { $mag <= 0 } { return $dmg }
  return [expr2 2 $dmg + 0.8 * $dmg * $mag / 50]
}

###  id    
proc mc:magic:correctid {mid id} {
  global spells_alias
  if { [mc:magic:exist $mid] } { return $mid }
  if { [info exist spells_alias($mid)] } { return $spells_alias($mid) }
  set nmid "[string index $mid 0][mc:getstat prof $id][string index $mid 1]"
  if { [mc:magic:exist $nmid] } { return $nmid }
  if { [info exist spells_alias($nmid)] } { return $spells_alias($nmid) }
  return $mid
}

proc mc:magic:exist {mg} {
  global spells

  return [info exist spells($mg)]
}

proc mc:magic:getstat {stt mg {lvl 0}} {
  global spells

  if {[info exist spells($mg)]} {
    array set mgrec $spells($mg)
    if {[info exist mgrec($stt)]} {
      if {$lvl == "*"} {
        return $mgrec($stt)
      } elseif {[llength $mgrec($stt)] < $lvl} {
        return [lindex $mgrec($stt) end]
      } {
        return [lindex $mgrec($stt) $lvl]
      }
    }
  }

  putlog "Error in <magic:getstat> $mg $stt $lvl"
  return 0
}

proc mc:magic:list {} {
  global spells
  return [array names spells]
}

proc mc:magic:xgetstat {stt mg lvl} {
  set ret [mc:magic:getstat $stt $mg $lvl]
  return $ret
}

#!! obsolete
proc mc:magic:getcircle {mgid} {
  global spells
  if {[info exist spells($mgid)]} {
    return [mc:magic:getstat circle $mgid]
  } {
    return [string index $mgid 0]
  }
}

proc mc:magic:getprof {mgid} {
  global spells
  if {[info exist spells($mgid)]} {
    return [mc:magic:getstat prof $mgid]
  } {
    return [string index $mgid 1]
  }
}

proc mc:magic:getname {mgid} {
  global spells
  return [mc:magic:getstat name $mgid *]
}

proc mc:magic:getmana {mgid id {llev "-"}} {
  global spells

  if {$llev == "-"} {
    set llev 0
    foreach p [mc:getstat magic $id] {
      if {[lindex $p 0] == $mgid} {
        set llev [expr [lindex $p 1] - 1]
        break
      }
    }
  }

  return [expr2 1 [mc:magic:getstat basemana $mgid $llev] * [mc:getstat manaprc $id]]
}
#!! end obsolete

proc mc:magic:xgetver {mgrec} {
  global spells
  global level type id targ aprc attstr prtstr time long name
  mc:magic:setglobal $mgrec

  if {[info exist spells($type)]} {
    if {[expr $aprc * 100] < [mc:magic:getstat minprc $type $level]} {
      return -1
    }
    return [mcmg:$type:ver]
  }
  putlog "Error in <magic:xgetver> $mgrec"
  return 0
}

proc mc:magic:xgetlong {mgrec} {
  global type level id
  mc:magic:setglobal $mgrec
  if {![mc:magic:getstat long $type]} { return 1 }
  return [mcmg:$type:long $level $id]
}

proc mc:magic:setglobal {mgrec} {
  global id targ attnick trgnick level
  global aprc type level attstr prtstr time long uid name

  array set mag $mgrec
  set id      $mag(id)
  set targ    $mag(targ)
  set attnick [ms:dfl $mag(attnick)]
  set trgnick [ms:dfl $mag(trgnick)]
  set level   $mag(level)
  set aprc    $mag(prc)
  set type    $mag(type)
  set level   $mag(level)
  set attstr  $mag(attstr)
  set prtstr  $mag(prtstr)
  set time    $mag(time)
  set name    $mag(name)
  if {[info exist mag(long)]} { set long    $mag(long) }
  if {[info exist mag(uid)] } { set uid     $mag(uid)  }
}

proc mc:magic:runspell {mgrec} {
  global id targ attnick trgnick level nnick param rundata
  global time long aprc type level attstr prtstr time long uid exp

  mc:magic:setglobal $mgrec

  if {$time == 0} {
    lappend param($targ) "$id $aprc $targ $type $level $attstr $time $long $uid"
  }

  incr time
  mcmg:$type:exp
  set ret [mcmg:$type:run]

  if {$ret != ""} {
    set rdid "${type}@${id}@${targ}@${long}"
    if {![array exist rundata] || ![info exist rundata($rdid)]} {
      set rundata($rdid) ""
    }
    lappend rundata($rdid) "$time $exp $ret"
  }

  return $ret
}

proc mc:magic:load {} {
  global mgload mc lstr errstr spells spells_alias

  putlog "Loading magic..."
  set fnl [glob -directory $mc(file.mage) "*.magic"]
  if {[array exist spells]} { unset spells }
  array_init spells
  array_init spells_alias
  set spellparam "postmode long targ no-self no-team no-other minprc"
  foreach lstr $fnl {
    if {[catch {uplevel #0 {source $lstr}} errstr]} {
      putlog "Error while load magic ${lstr}: $errstr"
    } else {
      array set mgload [spell_init]
      if {[info exist mgload(flags)]} {
        if {[llength [split $mgload(flags) ,]]>1} {
          set mgload(flags) [split $mgload(flags) ,]
        }
        foreach param $mgload(flags) {
          set mgload($param) 1
        }
        unset mgload(flags)
      }
      foreach param $spellparam {
        if {![info exist mgload($param)]} {
          set mgload($param) 0
        }
      }
      if { [info exist mgload(alias)] } {
        foreach alias $mgload(alias) {
          set spells_alias($alias) $mgload(code)
        }
      }
      if {[info exist spells($mgload(code))]} {
        putlog "Warning! Duplicate mage: $lstr"
      } {
        set spells($mgload(code)) [array get mgload]
      }
    }
    unset mgload
  }

  set mc(magupdate) 1
}

proc mc:magic:runpostspell {} {
  global spells

  foreach spell [array names spells] {
    if {[mc:magic:getstat postmode $spell]} {
      mcmg:$spell:post
    }
  }
}


proc mc:magic:flushfail {} {
  global ctm mgfail ct

  if {![array exist mgfail]} { return }
  foreach rec [array names mgfail] {
    set times [lsort -integer $mgfail($rec)]
    set rec [split $rec @]

    set type [lindex $rec 0]
    set id   [lindex $rec 1]
    set targ [lindex $rec 2]
    set long [lindex $rec 3]
    set why  ""
    switch -- [lindex $rec 4] {
      1 { set why $ct(41h) }
      2 { set why $ct(41i) }
      4 { set why $ct(41f) }
    }

    set prefix ""
    if { [llength $times] == 1 && [lindex $times 0] == 1 } {
      if {[info exist ctm(${type}.fail0)]} {
         mc:say [xformat3 $ctm(${type}.fail0) [mc:sex $id $targ] [ms:dfl [mc:getstat nick $id]] [ms:dfl [mc:getstat nick $targ]] $why]
       } {
         mc:say [xformat3 $ct(41) [mc:sex $id] [ms:dfl [mc:getstat nick $id]] [mc:magic:getname $type] [ms:dfl [mc:getstat nick $targ]] $why]
       }
    } {
      set prefix "\[[join $times ,]  $long\] "
      if {[info exist ctm(${type}.fail1)]} {
        mc:say "$prefix[xformat3 $ctm(${type}.fail1) [mc:sex $id $targ] [ms:dfl [mc:getstat nick $id]] [ms:dfl [mc:getstat nick $targ]] $why]"
      } {
        mc:say "$prefix[xformat3 $ct(41) [mc:sex $id] [ms:dfl [mc:getstat nick $id]] [mc:magic:getname $type] [ms:dfl [mc:getstat nick $targ]] $why]"
      }
    }
  }

  unset mgfail
}

proc mc:magic:say {txt} {
  global prefix
  global d_result

  if {[info exist prefix] && $prefix != ""} { set txt "$prefix $txt";set prefix "" }
  if {[info exist d_result]} {
    set txt "$txt $d_result"
    unset d_result
  }
  mc:say $txt
}

proc mc:magic:flushlong {} {
  global rundata
  global prefix data type id targ attnick trgnick exp datanum attstr prtstr

  if {![array exist rundata]} { return }
  foreach rec [array names rundata] {

     set xdata $rundata($rec)
     set data ""
     set time ""
     set exp 0
     set i   1
     set datanum 0
     foreach xd $xdata {
       set exp [expr $exp + [lindex $xd 1]]
       set ttm [lindex $xd 0]
       if {[llength $xd] > 2} {
         set xtime([lindex $xd 0].$i) [lrange $xd 2 end]
       } {
         set xtime([lindex $xd 0].$i) ""
       }
       incr i
     }
     foreach tm [lsort -real [array names xtime]] {
       lappend time [lindex [split $tm .] 0]
       lappend data $xtime($tm)
       if { [isrealnumber $xtime($tm)] } {
         set datanum [expr 1.0 * $datanum + $xtime($tm)]
       }
     }
     set time [lsort -integer $time]

     set rec [split $rec @]
     set type [lindex $rec 0]
     if { "[info procs mcmg:$type:postrun]" == "" } { continue }
     set id   [lindex $rec 1]
     set targ [lindex $rec 2]
     set long [lindex $rec 3]
     set attnick [ms:dfl [mc:getstat nick $id]]
     set trgnick [ms:dfl [mc:getstat nick $targ]]
     set attstr  [mc:round:var magatt $id]
     set prtstr  [mc:round:var magprt $id]
     set prefix ""
     if { [llength $time] != 1 || $long != 1} {
       set prefix "14\[12[join $time 14,12] 1412 $long14\]"
     }
     mcmg:$type:postrun

     unset data
     unset prefix
     unset time
     unset xtime
     unset targ
  }

  unset rundata
}


proc mc:magic:flushrun {} {
  global rundata

  global prefix data type id targ attnick trgnick exp datanum attstr prtstr datanum

  if {![array exist rundata]} { return }

  foreach rec [array names rundata] {
     set type [lindex [split $rec @] 0]
     if {[info procs mcmg:$type:flushrun] == ""} { continue }

     set xdata $rundata($rec)
     set data ""
     set time ""
     set exp 0
     set i   1
     set datanum 0
     foreach xd $xdata {
       set exp [expr $exp + [lindex $xd 1]]
       if {[llength $xd] > 2} {
         set xtime([lindex $xd 0].$i) [lrange $xd 2 end]
       } {
         set xtime([lindex $xd 0].$i) ""
       }
       incr i
     }
     foreach tm [lsort -real [array names xtime]] {
       lappend time [lindex [split $tm .] 0]
       lappend data $xtime($tm)
       if { [isrealnumber $xtime($tm)] } {
         set datanum [expr 1.0 * $datanum + $xtime($tm)]
       }
     }
     set time [lsort -integer $time]

     set rec [split $rec @]
     set id   [lindex $rec 1]
     set targ [lindex $rec 2]
     set long [lindex $rec 3]
     set attnick [mc:getstat nick $id]
     set trgnick [mc:getstat nick $targ]
     set attstr  [mc:round:var magatt $id]
     set prtstr  [mc:round:var magprt $id]
     set prefix ""
     if { [llength $time] != 1 || $long != 1} {
       set prefix "14\[12[join $time 14,12] 1412 $long14\]"
     }

     mcmg:$type:flushrun

     unset data
     unset prefix
     unset time
     unset xtime
  }
}

proc mc:magic:round_init {} {
  global crazy_arr

  if { [array exist crazy_arr] } { unset crazy_arr }
}

proc mc:magic:round_done {} {
  global crazy_arr

  if { [array exist crazy_arr] } { unset crazy_arr }
}

proc mc:mmag {n uh h t} {
    global mcs mc ct cth mcr cth ct mcc mce ts tl twis ti tv th ta tg te todo liv param nnick

    set id [mc:makeid $n]
    if {![mc:check $id -]} { return }
    if {![isnumber [set t [lindex [split $t] 1]]] || [set t [expr round($t)]] < 1} { set t "" }
    mc:magic:checkempty $id
    set pr [mc:getstat magic $id]

    foreach p $pr {
        if {[lindex $p 1] == "-"} {
            mc:msg $n $cth(27)
            return
        }
        set mid      [lindex $p 0]
        set loclevel [lindex $p 1]
        set magmana [mc:magic:getmana $mid $id [expr $loclevel - 1]]
        set magname [mc:magic:getname $mid]
        set magtype [mc:magic:getstat  type     $mid]
        set llev    [mc:magic:getstat  circle   $mid]
        if {$t != "" && $llev != $t} { continue }

        if {![array exist mmma] || ![info exist mmma($llev)]} {
          set mmma($llev) ""
        }
        lappend mmma($llev) [format $cth(28) $mid [format $cth(28$magtype) $magname] $magmana $loclevel]
    }
    if {![array exist mmma]} {
      mc:msg $n [format $cth(27a) $t]
      return
    }
    for {set i 9} {$i >= 1} {incr i -1} {
        if {[info exist mmma($i)]} { mc:msg $n [format $cth(29) $i [join $mmma($i) ", "]] }
    }
}

proc mc:mmagexplain {n uh h t} {
    global mcs mc ct cth mcr cth ct mcc mce ts tl twis ti tv th ta tg te todo liv param

    set id [mc:makeid $n]
    if {![mc:check $id -]} { return }

    if {[llength $t] < 2} {
        mc:msg $n $cth(magexf)
        return
    }
    set mx [lindex $t 1]
    set mx [mc:magic:correctid $mx $id]
    if {![mc:magic:exist $mx]} {
        mc:msg $n [format $cth(magexn) $mx]
        return
    }

   set out ""
   set desc [mc:magic:getstat desc $mx *]
   set name [mc:magic:getname $mx]
   if {![mc:magic:getstat targ $mx]} {
     mc:msg $n [format $cth(magexdn) $name $desc]
   } {
     mc:msg $n [format $cth(magexd) $name $desc]
     if {![mc:magic:getstat no-self  $mx]} { set out [lappend out $cth(magexz1)] }
     if {![mc:magic:getstat no-team  $mx]} { set out [lappend out $cth(magexz2)] }
     if {![mc:magic:getstat no-other $mx]} { set out [lappend out $cth(magexz3)] }
   }

   ###    
   if {$out != ""} {
     set out [join $out ", "]
     mc:msg $n [format $cth(magexz) $out]
   }

}

proc mc:mlearn {n uh h t} {
    global mcs mc ct cth mcr cth ct mcc mce ts tl twis ti tv th ta tg te todo liv param

    set id [mc:makeid $n]
    if { ![mc:check $id -] } { return }
    mc:magic:checkempty $id

    set pr [mc:getstat magic $id]
    if {[llength $t] < 2} {
        mc:msg $n $cth(35)
        return
    }
    set lev [lindex $t 1]
    if {![isnumber $lev] || $lev < 1} { [uerror [format $cth(36) $lev]] }
    set prlev [mc:getstat level $id]
    if {[expr ($lev - 1) * 2 + 1] > $prlev} { [uerror $cth(37)] }
#    if {$lev > [mc:getstat bonus $id]}
    if {1 > [mc:getstat bonus $id]} { [uerror $cth(38)] }

    set magavail ""
    foreach mgg [mc:magic:list] {
      if {[mc:magic:getcircle $mgg] == $lev && [mc:checkin [mc:getstat prof $id] [mc:magic:getprof $mgg]]} {
        set loclvl 0
        foreach umgg $pr {
          if { [lindex $umgg 0] == $mgg } {
            set loclvl [lindex $umgg 1]
            break
          }
        }
        if { $loclvl >= [mc:magic:getstat levels $mgg] } { continue }
        lappend magavail "$mgg $loclvl"
      }
    }
    if {$magavail == ""} { [uerror [format $cth(39) $lev]] }
    #set ver [expr ($prlev - $lev + 3) * 25]
    # (<_> - <> + 3) * 25
    set ver [expr ($prlev - (($lev - 1) * 2 + 1) + 3) * 25]
    # +5%   .
    set ver [expr $ver + 5 * [mc:getstat remort $id]]

    if {$ver > 100} { set ver 100 }

    set mrand [mc:die 1d100]
    mc:setstat bonus $id [expr [mc:getstat bonus $id] - 1]
    if {$mrand > $ver} { [uerror $cth(40)] }

    set mgg [lrandom $magavail]
    set mid [lindex $mgg 0]
    set mlv [lindex $mgg 1]
    incr mlv

    set name [mc:magic:getname $mid]
    if { $mlv == 1 } {
      mc:msg $n [format $cth(42) $name]
      lappend pr "$mid $mlv"
    } {
      mc:msg $n [format $cth(43) $name $mlv]
      set newpr ""
      foreach mrec $pr {
        if { [lindex $mrec 0] == $mid } { set mrec [lreplace $mrec 1 1 $mlv] }
        lappend newpr $mrec
      }
      set pr $newpr
    }

    mc:setstat magic $id [lsort -ascii -index 0 $pr]
}

proc mc:magic:checkempty {id} {
  set newmag ""
  set bonus  0
  set prof   [mc:getstat prof $id]
  foreach mrec [mc:getstat magic $id] {
    if { [lindex $mrec 0] == "-" } { continue }
    if { [mc:checkin $prof [mc:magic:getprof [lindex $mrec 0]]] } {
      lappend newmag $mrec
    } {
      incr bonus [lindex $mrec 1]
    }
  }
  if { $bonus != 0 } {
    mc:setstat magic $id $newmag
    mc:setstat bonus $id [expr [mc:getstat bonus $id] + $bonus]
  }
}

###   ,  
proc mc:magic:autoupdate {} {
  global mc spells

  if {![info exist mc(magupdate)]} { return 0 }
  unset mc(magupdate)

  foreach id [mc:getalluser] {
    set nmag ""
    set upd  0
    foreach mrec [mc:getstat magic $id] {
      if { [lindex $mrec 0] != "-" && ![info exist spells([lindex $mrec 0])] } {
        mc:setstat bonus $id [expr [mc:getstat bonus $id] + [lindex $mrec 1]]
        set magdel([lindex $mrec 0]) 1
        set upd 1
      } {
        lappend nmag $mrec
      }
    }
    if { $upd } { mc:setstat magic $id $nmag }
  }

  if { [info exist magdel] } {
    putlog "[grn]MagAutoUpdate[hblk]:[cl] Deleted nonexists magic [hblk]\([red][array names magdel][hblk]\)[cl]."
  }
  return 1
}

###    ,    .
proc mc:magic:roundautoupdate {} {
  global spells

  if {![mc:magic:autoupdate]} { return }

  foreach id [mc:game:livlist] {
    set ndo ""
    set upd 0
    foreach dodo [mc:round:get_todo $id] {
      if { [lindex $dodo 0] == "m" && ![info exist spells([lindex $dodo 3])] } {
        putlog "[grn]MagAutoUpdate[hblk]:[cl] Deleted nonexists magic [hblk]\([red][lindex $dodo 3][hblk]\)[cl] from [cyn]$id[cl] order."
        set upd 1
      } {
        lappend ndo $dodo
      }
    }
    if { $upd } { mc:round:set_todo $id $ndo }
  }
}